/* APE (Allelic Path Explorer) v.1.0
   searches for a consistent genotypic configuration 
   when the known pedigree structure 
   and genetic data for some individuals are given. 


   Copyright (C) Matti Pirinen
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   version 2 as published by the Free Software Foundation.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program (gpl.txt); if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.

   www.gnu.org/licenses/gpl.html
*/

/* FILE: APE_order.c
   Contains functions that are used in the ordering process of APE.
   For list of functions see APE.h
*/




int compar2(const void *A, const void *B)
{

  t_list *a = (t_list*) A;
  t_list *b = (t_list*) B;

  if(a->proba < b->proba)
    return 1;
  else 
  return -1;

}



int compare_group(const void *A, const void *B)
{

  o_list *a = (o_list*) A;
  o_list *b = (o_list*) B;

  if(b->score > a->score )
    return 1;
     else
   return -1;

}

int count_different_genes(pair genes, pair genes2)
{

  int sum=0;
  int undefined;



  if( (genes[0] == 0 && genes[1] == 0) || (genes2[0] == 0 && genes2[1] == 0) )
    return 0;

  undefined= (genes[0] == 0) + (genes[1] == 0) + (genes2[0] == 0) + (genes2[1] == 0);

  sum += (genes[0] > 0 && genes[0] != genes2[0] && genes[0] != genes2[1]);
  sum += (genes[1] > 0 && genes[1] != genes2[0] && genes[1] != genes2[1]);
  sum += (genes2[0] > 0 && genes2[0] != genes[0] && genes2[0] != genes[1]);
  sum += (genes2[1] > 0 && genes2[1] != genes[0] && genes2[1] != genes[1]);

  if(sum==0)
    return sum;

   return sum+(4-2*undefined);

}

double diff_component(pair *tgenes, COMP *C, int g,int c, double ***kinship)
{

  int i,j,ind,ind2,pind,pind2;
  double sum=0.0;

  for(i=0;i<C[g].components[c].nparents;i++)
    {
      ind=C[g].components[c].parents[i];
      pind=C[g].components[c].pind[i];

      for(j=i+1;j<C[g].components[c].nparents;j++)
	{
	  ind2=C[g].components[c].parents[j];
	  pind2=C[g].components[c].pind[j];

	  if(diff_exp!=1.0)
	    sum +=  kinship[g][pind][pind2]*pow(count_different_genes(tgenes[ind],tgenes[ind2])*1.0/8.0,diff_exp); 
	  else
	    sum +=  kinship[g][pind][pind2]*count_different_genes(tgenes[ind],tgenes[ind2])/8.0;
	}

    }
  
  return sum;
}



double diff_group_to_component(pair *tgenes, COMP *C, int g,int ngroup, double ***kinship)
{

  int i,j,k,c,cc,ind,ind2,pind,pind2;
  double sum=0.0;

  c=C[g].order[ngroup];

  for(k=0;k<ngroup;k++)
    {
      cc=C[g].order[k];
      for(i=0;i<C[g].components[cc].nparents;i++)
	{
	  ind=C[g].components[cc].parents[i];
	  pind=C[g].components[cc].pind[i];

	  for(j=0;j<C[g].components[c].nparents;j++)
	    {
	      ind2=C[g].components[c].parents[j];
	      pind2=C[g].components[c].pind[j];

	      if(diff_exp!=1.0)
		sum +=  kinship[g][pind][pind2]*pow(count_different_genes(tgenes[ind],tgenes[ind2])*1.0/8.0,diff_exp); 
	      else
		sum += kinship[g][pind][pind2]*count_different_genes(tgenes[ind],tgenes[ind2])/8.0; 
	    }

	}
    }
  
  return sum;
}


void find_component_ordering(COMP *C, double ***kinship)
{

  /* Finds an order according to which components will be merged
     to cluster at the ordering procedure.
     Ordering is a part of the COMP structure.*/

  int i,ii,j,l,c,cc,maxi,pind,pind2;
  int *available;
  double **sums,max;
  double *available_sums;
  pair ends;

  for(l=0;l<N_levels-1;l++)
    {

      if( (C[l].order=(int*) malloc(C[l].ncomponents*sizeof(int))) == NULL)
	alloc_error("find_component_ordering\n");

      if(C[l].ncomponents<3)
	{
	  for(i=0;i<C[l].ncomponents;i++)
	    C[l].order[i]=i;
	  continue; //loop with variable l
	}

      if( (sums=(double**) calloc(C[l].ncomponents,sizeof(double*)))==NULL)
	alloc_error("find_component_ordering with 'sums'\n");


      for(c=0;c<C[l].ncomponents;c++)
	{
	  if( (sums[c]=(double*) calloc(C[l].ncomponents,sizeof(double)))==NULL)
	    alloc_error("find_component_ordering with 'sums'\n");
	}

      max=-1.0;
      ends[0]=ends[1]=1;

      for(c=0;c<C[l].ncomponents;c++)
	{
	  for(i=0;i<C[l].components[c].nparents;i++)
	    for(ii=i+1;ii<C[l].components[c].nparents;ii++)
	      {
		pind=C[l].components[c].pind[i];
		pind2=C[l].components[c].pind[ii];
		sums[c][c] += kinship[l][pind][pind2];
	      }

	  for(cc=0;cc<c;cc++)
	    {
	      for(i=0;i<C[l].components[c].nparents;i++)
		for(ii=0;ii<C[l].components[cc].nparents;ii++)
		  {
		    pind=C[l].components[c].pind[i];
		    pind2=C[l].components[cc].pind[ii];
		    sums[c][cc] += kinship[l][pind][pind2];
		  }

	      if(sums[c][cc] + sums[c][c] + sums[cc][cc]> max)
		{
		  max=sums[c][cc]+ sums[c][c] + sums[cc][cc];
		  ends[0]=cc;
		  ends[1]=c;
		}

	      sums[cc][c]=sums[c][cc];
	    }
	}

      /*printf("Level %d, %d components:\n",l,C[l].ncomponents);
      if(l<N_levels-2)
	for(c=0;c<C[l].ncomponents;c++)
	  {
	    for(cc=0;cc<C[l].ncomponents;cc++)
	      printf("%5.3lf ",sums[c][cc]); 
	    printf("\n");
	  }
	  printf("Ordering:\n");*/
     
      C[l].order[0]=ends[0];
      C[l].order[1]=ends[1];

      if( (available_sums=(double*) calloc(C[l].ncomponents-2,sizeof(double)))==NULL)
	alloc_error("find_component_ordering with 'available_sums'\n");

      if( (available=(int*) malloc((C[l].ncomponents-2)*sizeof(int)))==NULL)
	alloc_error("find_component_ordering with 'available'\n");

      max=sums[ends[0]][ends[0]]+sums[ends[1]][ends[1]];
      for(i=0,j=0;i<C[l].ncomponents;i++)
	if(i != ends[0] && i != ends[1])
	  {
	    available[j]=i;
	    available_sums[j++]=max+sums[i][i]+sums[i][ends[0]]+sums[i][ends[1]];
	  }
      
      for(i=C[l].ncomponents-2;i>0;i--)
	{
	  maxi=0;
	  max=0.0;

	  for(j=0;j<i;j++)
	    {
	      if(max<available_sums[j])
		{
		  max=available_sums[j];
		  maxi=j;
		}
	    }
	  C[l].order[C[l].ncomponents-i]=available[maxi];

	  for(j=0;j<i;j++)
	    available_sums[j] += sums[maxi][maxi]+sums[maxi][available[j]];

	  available[maxi]=available[i-1];
	  available_sums[maxi]=available_sums[i-1];
	}

      for(c=0;c<C[l].ncomponents;c++)
	{
	  //printf("%d ",C[l].order[c]);
	  
	  for(j=0,ii=0;j<C[l].ncomponents;j++)
	    if(C[l].order[j]==c)
	      ii=1;

	  if(!ii)
	    {
	      printf("ERROR with find_component_ordering!\n");
	      exit(6);
	    }
	}
      //printf("\n");

      for(c=0;c<C[l].ncomponents;c++)
	free(sums[c]);
      free(sums);
      free(available_sums);
      free(available);
    }
      

}



void order_component(pair **ttensor,double *proba, int ntensor)
{

  int i;
  t_list *list;

  if( (list = (t_list*) malloc(ntensor*sizeof(t_list)))==NULL)
     alloc_error("order_component\n");

  for(i=0;i<ntensor;i++)
    {
      list[i].tensor=ttensor[i];
      list[i].proba=proba[i];
    }

  qsort(list, ntensor, sizeof(t_list), *compar2);

  for(i=0;i<ntensor;i++)
    {
      ttensor[i]=list[i].tensor;
      proba[i]=list[i].proba;
    }

  free(list);

}


void order_generation(g_list *G, COMP *C, int g, double ***kinship, pair* tgenes, double *ratio_known)
{

  int **group,**tgroup;
  double *proba;
  double *difference;
  double sum,max;
  int c,ngroup,ncomp;
  o_list *list;
  int i,ci,j,k,cc,current,ind;

  printf("%6.0g -> ",exp(G[g].log_n));

  c=C[g].order[0];
  
  ngroup = G[g].nconfig[c];

  if( (group = (int**) malloc(ngroup*sizeof(int*))) == NULL)
    alloc_error("order_generation with 'group'");

  for(i=0;i<ngroup;i++)
    if( (group[i] = (int*) malloc(sizeof(int))) == NULL)
      alloc_error("order_generation with 'group'");

  if( (difference = (double*) malloc(ngroup*sizeof(double))) == NULL)
    alloc_error("order_generation with 'difference'");
  
  if( (proba = (double*) malloc(ngroup*sizeof(double))) == NULL)
    alloc_error("order_generation with 'proba'");

  for(i=0;i<ngroup;i++)
    {
      group[i][0]=i;
      proba[i]=G[g].comp[c].proba[i];
      for(j=0;j<C[g].components[c].nparents;j++)
	{
	  ind=C[g].components[c].parents[j];
	  tgenes[ind][0]=G[g].comp[c].tensor[i][j][0];
	  tgenes[ind][1]=G[g].comp[c].tensor[i][j][1];
	}
      difference[i]=diff_component(tgenes,C,g,c,kinship);
    }

  for(ci=1;ci<C[g].ncomponents;ci++)
    {
      c=C[g].order[ci];
      ncomp = G[g].nconfig[c];
      current=0;
      sum=0.0;
      max=0.0;

      if( (list = (o_list*) malloc(ngroup*G[g].nconfig[c]*sizeof(o_list)))==NULL)
	alloc_error("order_generation with 'list'");

      for(i=0;i<ngroup;i++)
	{
	  for(j=0;j<ci;j++)
	    {
	      cc=C[g].order[j];
	      for(k=0;k<C[g].components[cc].nparents;k++)
		{
		  ind=C[g].components[cc].parents[k];
		  tgenes[ind][0]=G[g].comp[cc].tensor[group[i][j]][k][0];
		  tgenes[ind][1]=G[g].comp[cc].tensor[group[i][j]][k][1];
		} 
	    }

	  for(j=0;j<ncomp;j++)
	    {
	      for(k=0;k<C[g].components[c].nparents;k++)
		{
		  ind=C[g].components[c].parents[k];
		  tgenes[ind][0]=G[g].comp[c].tensor[j][k][0];
		  tgenes[ind][1]=G[g].comp[c].tensor[j][k][1];
		}

	      list[current].index[0]=i;
	      list[current].index[1]=j;
	      list[current].proba=proba[i]*G[g].comp[c].proba[j];
	      list[current].difference=difference[i]+diff_component(tgenes,C,g,c,kinship)+diff_group_to_component(tgenes,C,g,ci,kinship);
	      /*if(list[current].difference>max)
		max=list[current].difference;*/
	      ++current;
	    }
	}

      current=0;

      /*if(sum<1e-30) sum=1.0;
	min=1.0;*/
      for(i=0;i<ngroup;i++)
	for(j=0;j<ncomp;j++)
	  {
	    list[current].score=log(list[current].proba)-kappa*list[current].difference;
	    ++current;
	  }

	 /*if( (list[current++].score=list[current].proba*exp(-kappa*(1.0-ratio_known[g])*list[current].difference)) < min)
	      min=list[current-1].score;*/
      //list[current++].score=list[current].proba*exp(-kappa*list[current].difference);
	    //list[current++].score=list[current].proba*exp(-kappa*(max/(max-0.5*list[current].difference)-1));
      /*if(min<=0.0)
      	printf("\nWARNING: underflow in ordering! (level=%d,component=%d)\n",g,c);*/

      qsort(list, ngroup*ncomp, sizeof(o_list), *compare_group);

      /*printf("Level %d, component %d\n",g,ci);
      printf("GROUP:\n");
      for(i=0;i<ngroup;i++)
	{	
	  printf("%d:",i);
	  for(j=0;j<ci;j++)
	    printf("%d ",group[i][j]);
	  printf("\n");
	  }
      printf("LIST:\n");
      for(i=0;i<ngroup*ncomp;i++)
	{
	  printf("%d %d ",list[i].index[0],list[i].index[1]);
	  printf("%6.3lf %6.3lf %6.3lf ",list[i].proba,list[i].difference,list[i].score);
	  printf("%10.3g\n", exp(list[i].score-list[0].score));
	  }*/
      
      i=ngroup*ncomp;
      k=ngroup;

      ngroup=(M_clu>i || !M_clu)?i:M_clu;

      if( (tgroup=(int**) malloc(ngroup*sizeof(int*)))==NULL )
	alloc_error("order_generation with 'tgroup'");

      for(i=0;i<ngroup;i++)
	 if( (tgroup[i]=(int*) malloc((ci+1)*sizeof(int)))==NULL )
	   alloc_error("order_generation with 'tgroup'");

      if( (difference=(double*) realloc(difference,ngroup*sizeof(double))) == NULL)
	alloc_error("order_generation with 'difference'");

      if( (proba=(double*) realloc(proba,ngroup*sizeof(double))) == NULL)
	alloc_error("order_generation with 'proba'");

      sum=0.0;
      for(i=0;i<ngroup;i++)
	{
	  for(j=0;j<ci;j++)
	    tgroup[i][j]=group[list[i].index[0]][j];
	  tgroup[i][ci]=list[i].index[1];

	  sum += list[i].proba;
	}

      for(i=0;i<ngroup;i++)
	{
	  proba[i]=list[i].proba/sum;
	  difference[i]=list[i].difference;
	}

      free(list);
      list=NULL;
      for(i=0;i<k;i++)
	free(group[i]);
      free(group);

      group=tgroup;
    }

  if(M_lev > 0 && M_lev < ngroup)
    {
      for(i=M_lev;(unsigned) i<ngroup; i++)
	free(group[i]);
      
      G[g].n = M_lev;
    }
  else
    G[g].n=ngroup;

  printf(" %ld\n",G[g].n);

  if( (G[g].ordered = (config *) malloc(G[g].n*sizeof(config))) == NULL )
    alloc_error("order_generation with 'G.ordered'");

  for(i=0;i<G[g].n;i++)
    {
      if( (G[g].ordered[i].component = (int*) malloc(C[g].ncomponents*sizeof(int))) == NULL )
	alloc_error("order_generation with 'G.ordered.component'");
      for(j=0;j<C[g].ncomponents;j++)
	G[g].ordered[i].component[C[g].order[j]]=group[i][j];
      free(group[i]);
    }

  if(G[g].n==0)
    {
      printf("\nNo configurations with positive probability. Program aborted.\n");
      exit(6);
    }

  free(group);
  free(proba);
  free(difference);

  
}



